/*!
 * \file
 *
 *  Webserver Source File
 *
 *  Webserver hardware is a MicroChip ENC28J60.
 *  datasheet:      ethernet datasheet
 *
 */
#if AVR
#include "__NetEthEnc28j60.h"
#define SOCKET_Dsc SOCKET_28j60_Dsc
#define UDP_Dsc UDP_28j60_Dsc
#define Net_Ethernet_stackInitTCP                             Net_Ethernet_28j60_stackInitTCP
#define Net_Ethernet_Init(myMacAddr, myIpAddr, DUPLEX)        Net_Ethernet_28j60_Init(myMacAddr, myIpAddr, DUPLEX)
#define Net_Ethernet_confNetwork(ipMask, gwIpAddr, dnsIpAddr) Net_Ethernet_28j60_confNetwork(ipMask, gwIpAddr, dnsIpAddr)
#define Net_Ethernet_doPacket                                 Net_Ethernet_28j60_doPacket
#define Net_Ethernet_getByte                                  Net_Ethernet_28j60_getByte
#define Net_Ethernet_getBytes                                 Net_Ethernet_28j60_getBytes
#define Net_Ethernet_UserTCP                                  Net_Ethernet_28j60_UserTCP
#define Net_Ethernet_putStringTCP                             Net_Ethernet_28j60_putStringTCP
#define Net_Ethernet_putConstStringTCP                        Net_Ethernet_28j60_putConstStringTCP
#define Net_Ethernet_putByteTCP                               Net_Ethernet_28j60_putByteTCP
#define Net_Ethernet_putBytesTCP                              Net_Ethernet_28j60_putBytesTCP
#define Net_Ethernet_putConstByteTCP                          Net_Ethernet_28j60_putConstByteTCP
#define Net_Ethernet_putConstBytesTCP                         Net_Ethernet_28j60_putConstBytesTCP
#define Net_Ethernet_bufferEmptyTCP                           Net_Ethernet_28j60_bufferEmptyTCP
#define Net_Ethernet_UserUDP                                  Net_Ethernet_28j60_UserUDP
#define Net_Ethernet_dnsResolve                               Net_Ethernet_28j60_dnsResolve
#define Net_Ethernet_sendUDP                                  Net_Ethernet_28j60_sendUDP
#define Net_Ethernet_connectTCP                               Net_Ethernet_28j60_connectTCP
#define Net_Ethernet_disconnectTCP                            Net_Ethernet_28j60_disconnectTCP
#define Net_Ethernet_startSendTCP                             Net_Ethernet_28j60_startSendTCP
#define Net_Ethernet_UserTimerSec                             Net_Ethernet_28j60_UserTimerSec
#define Net_Ethernet_bufferFreeSize                           Net_Ethernet_28j60_bufferFreeSize

#define packetVector IVT_ADDR_INT1

#define NUM_OF_SOCKET NUM_OF_SOCKET_28j60

#else

#include "__NetEthInternal.h"
#include "dma_mem.h"

#define SOCKET_Dsc SOCKET_Intern_Dsc
#define UDP_Dsc UDP_Intern_Dsc
#define Net_Ethernet_stackInitTCP                             Net_Ethernet_Intern_stackInitTCP
#define Net_Ethernet_Init                                     Net_Ethernet_Intern_Init
#define Net_Ethernet_confNetwork(ipMask, gwIpAddr, dnsIpAddr) Net_Ethernet_Intern_confNetwork(ipMask, gwIpAddr, dnsIpAddr)
#define Net_Ethernet_doPacket                                 Net_Ethernet_Intern_doPacket
#define Net_Ethernet_getByte                                  Net_Ethernet_Intern_getByte
#define Net_Ethernet_getBytes                                 Net_Ethernet_Intern_getBytes
#define Net_Ethernet_UserTCP                                  Net_Ethernet_Intern_UserTCP
#define Net_Ethernet_putStringTCP                             Net_Ethernet_Intern_putStringTCP
#define Net_Ethernet_putConstStringTCP                        Net_Ethernet_Intern_putConstStringTCP
#define Net_Ethernet_putByteTCP                               Net_Ethernet_Intern_putByteTCP
#define Net_Ethernet_putBytesTCP                              Net_Ethernet_Intern_putBytesTCP
#define Net_Ethernet_putConstByteTCP                          Net_Ethernet_Intern_putConstByteTCP
#define Net_Ethernet_putConstBytesTCP                         Net_Ethernet_Intern_putConstBytesTCP
#define Net_Ethernet_bufferEmptyTCP                           Net_Ethernet_Intern_bufferEmptyTCP
#define Net_Ethernet_UserUDP                                  Net_Ethernet_Intern_UserUDP
#define Net_Ethernet_dnsResolve                               Net_Ethernet_Intern_dnsResolve
#define Net_Ethernet_sendUDP                                  Net_Ethernet_Intern_sendUDP
#define Net_Ethernet_connectTCP                               Net_Ethernet_Intern_connectTCP
#define Net_Ethernet_disconnectTCP                            Net_Ethernet_Intern_disconnectTCP
#define Net_Ethernet_startSendTCP                             Net_Ethernet_Intern_startSendTCP
#define Net_Ethernet_UserTimerSec                             Net_Ethernet_Intern_userTimerSec
#define Net_Ethernet_bufferFreeSize                           Net_Ethernet_Intern_bufferFreeSize

#define packetVector IVT_INT_ETH

#define NUM_OF_SOCKET NUM_OF_SOCKET_Intern

#include "bargraph.h"

#endif

#include "websvr.h"
#include "built_in.h"
#include "__Lib_FAT32.h"
#include "__Time.h"

#define FILENAMESIZE    13
#define DOMAINSIZE      50
#define LARGEBUFFERSIZE 512
#define GETPOSTSIZE     512
#define MAX_ALIVE_TIME  10
#define SWAPFILE "download.bin"

#define CR        0x0D
#define LF        0x0A

#define skipNext(buffer, c) while(*buffer && *buffer != c){buffer++;}if(*buffer){*buffer = '\0'; buffer++;}

#ifdef DEBUG
#define debug_print(str) UART1_Write_Text(str)
#else
#define debug_print(str) asm nop
#endif
/*!
 *  \brief States for Sockets
 *
 *  Socket states are represented in this enumerated type. The various stages of how a socket is
 *  parsed, validated, and served are representative in the members located in this structure.
 *
 *  \note
 *    As of current, not all modes are implemented.
 *
 */
typedef enum
{
    SM_HTTP_IDLE = 0,                       /*!< Socket is idle*/
    SM_HTTP_GET_REQUEST,                    /*!< Gets the full request from client */
    SM_HTTP_PARSE_REQUEST,                  /*!< Parses the first line for a file name and GET args*/
    SM_HTTP_FIND_FILE,                      /*!< Checks for file existance*/
    SM_HTTP_FILE_NOT_FOUND,                 /*!< requested file not found*/
    SM_HTTP_PARSE_HEADERS,                  /*!< Reads and parses headers one at a time*/
    SM_HTTP_AUTHENTICATE,                   /*!< Validates the current authorization state*/
    SM_HTTP_PROCESS_GET,                    /*!< Invokes user callback for GET args or cookies*/
    SM_HTTP_PROCESS_POST,                   /*!< Invokes user callback for POSTed data*/
    SM_HTTP_PROCESS_REQUEST,                /*!< Begins the process of returning data*/
    SM_HTTP_SERVE_HEADERS,                  /*!< Sends any required headers for the response*/
    SM_HTTP_SERVE_COOKIES,                  /*!< Adds any cookies to the response*/
    SM_HTTP_OPEN_FILE,                      /*!< OPens the requested file*/
    SM_HTTP_SERVE_BODY,                     /*!< Serves the actual content*/
    SM_HTTP_SERVE_BODY2,                    /*!< Serves the actual content */
    SM_HTTP_SEND_FROM_CALLBACK,             /*!< Invokes a dynamic variable callback*/
    SM_HTTP_WAIT_TILL_DONE,                 /*!< Waits until all bytes are served*/
    SM_HTTP_DISCONNECT,                     /*!< Disconnects the server and closes all files*/

    SM_HTTP_SVR_SEND_REQUEST,               /*!< */
    SM_HTTP_SVR_GET_RESPONSE,               /*!< */
    SM_HTTP_SVR_PARSE_RESPONSE,             /*!< */
    SM_HTTP_SVR_OPEN_FILE,                  /*!< */
    SM_HTTP_SVR_GET_DATA,                   /*!< */
    SM_HTTP_SVR_DOWNLOAD_COMPLETE,          /*!< */
    SM_HTTP_SVR_DISCONNECT                  /*!< */

} SM_SOCKET_SM;

// File type definitions
typedef enum
{
    HTTP_TXT = 0,            // File is a text document
    HTTP_HTM,                // File is HTML (extension .htm)
    HTTP_HTML,               // File is HTML (extension .html)
    HTTP_CGI,                // File is HTML (extension .cgi)
    HTTP_XML,                // File is XML (extension .xml)
    HTTP_CSS,                // File is stylesheet (extension .css)
    HTTP_GIF,                // File is GIF image (extension .gif)
    HTTP_PNG,                // File is PNG image (extension .png)
    HTTP_JPG,                // File is JPG image (extension .jpg)
    HTTP_JAVA,               // File is java (extension .java)
    HTTP_WAV,                // File is audio (extension .wav)
    HTTP_UNKNOWN             // File type is unknown
} HTTP_FILE_TYPE;

/*!
 *  \struct type of connection
 *
 */
typedef enum
{
    CONNECTION_KEEP_ALIVE = 0,
    CONNECTION_CLOSE

} HTTP_connection_type;

/*!
 *  \struct Encoding of connection
 *
 */
typedef enum
{
    ENCODING_TEXT = 0,
    ENCODING_CHUNKED,
    ENCODING_ZIP

} HTTP_transfer_encoding;

/*!
 *  \struct Response of SNTP Server
 *
 *  \brief Used to hold the return response from SNTP time server
 *
 */
typedef struct
{
    uint32_t epoch;          /*!< Time stamp from Jan1 1900 */
    uint8_t serverFlags;
    uint8_t serverStratum;
    int8_t serverPrecision;

} SNTP_Response_t;


/*!
 *  \struct Holds server request related variables
 *
 */
struct
{
    // Flags
    bool request_send_flag;
    // Ethernet Socket
    SOCKET_Dsc* eth_socket_download;
    // File length
    uint32_t file_length;
    // Capture length
    uint32_t capture_length;
    // Transmit length
    uint16_t transmit_size;
    // Position counter for chunked encoding
    uint32_t position;
    // Content length - may not be needed.
    uint16_t remaining;
    // Download file variables
    __HANDLE swap_handle;
    __HANDLE download_handle;
    __SECTOR swap_sector;

} server_request;

// Webpages
typedef struct
{
    // String representation of page
    char page_name[13];
    // Size of webpage in string
    char page_size_str[12];
    // Type of page ex. HTTP_HTML, HTTP_CSS, etc.
    HTTP_FILE_TYPE page_type;
    // Flag for static or dynamic
    bool static_page_flag;
    // If the page is a static page located in ROM pointer is here
    struct
    {
        const char* const start_ptr;
        const char* current_ptr;
        const char* const end_ptr;

    } static_page;
    // If the page is dynamic and will be located in RAM pointer here
    struct
    {
        char* const start_ptr;
        char* current_ptr;
        char* end_ptr;
        void ( *page_update )( void );

    } dynamic_page;

} webpage_t;

/*!
 * \struct socketNode_desc
 *
 * \brief Socket struct with file buffer information from SD card
 *
 *  \note Multiple socket_desc can be used to speed up transfers, however
 *  the memory requirement for this can grow quickly based on buffer sizes.
 */
typedef struct
{
    // Socket for TCP connection
    SOCKET_Dsc* eth_socket;
    // State of socket
    SM_SOCKET_SM socketSM;
    // Connection Type
    HTTP_connection_type connect_type;
    // Encoding Type
    HTTP_transfer_encoding encoding_type;
    // Socket buffers
    char buffer_large[LARGEBUFFERSIZE];
    // Buffer position
    uint16_t buffer_position;
    // Filename
    char file_name[FILENAMESIZE];
    // Alive time for persistent connections
    volatile uint8_t alive_time;
    // Webpage
    webpage_t* webpage;

} socket_t;


/*****************************
 *   Web Strings
 ****************************/
// File type extensions corresponding to HTTP_FILE_TYPE
static const code char* http_file_extensions[HTTP_UNKNOWN+1] =
{
    "txt",                       // HTTP_TXT
    "htm",                       // HTTP_HTM
    "html",                      // HTTP_HTML
    "cgi",                       // HTTP_CGI
    "xml",                       // HTTP_XML
    "css",                       // HTTP_CSS
    "gif",                       // HTTP_GIF
    "png",                       // HTTP_PNG
    "jpg",                       // HTTP_JPG
    "js",                        // HTTP_JAVA
    "wav",                       // HTTP_WAV
    "\0\0\0"                     // HTTP_UNKNOWN
};



// Content-type strings corresponding to HTTP_FILE_TYPE
static const code char* http_content_types[HTTP_UNKNOWN+1] =
{
    "text/plain\n\n",            // HTTP_TXT
    "text/html\n\n",             // HTTP_HTM
    "text/html\n\n",             // HTTP_HTML
    "text/html\n\n",             // HTTP_CGI
    "text/xml\n\n",              // HTTP_XML
    "text/css\n\n",              // HTTP_CSS
    "image/gif\n\n",             // HTTP_GIF
    "image/png\n\n",             // HTTP_PNG
    "image/jpeg\n\n",            // HTTP_JPG
    "application/x-javascript\n\n",  // HTTP_JAVA
    "audio/x-wave\n\n",          // HTTP_WAV
    ""                       // HTTP_UNKNOWN
};

static const code char* http_response_not_found = "HTTP/1.1 404 Not found\n\n";
static const code char* http_response_OK        = "HTTP/1.1 200 OK";
static const code char* http_response_OK_old    = "HTTP/1.0 200 OK";

static const code char* http_response_header    =
    "HTTP/1.1 200 OK\r\nServer: Alphaloewe/1.0\r\nConnection: keep-alive\r\nDate: ";
static const code char* http_response_length    = "\r\nContent-Length: ";
static const code char* http_response_type      = "\r\nContent-Type: ";

static const code char* http_request_host       = " HTTP/1.1\r\nHost: ";
static const code char* http_request_end        = "\r\nConnection: keep-alive\r\nAccept: */*\r\n\r\n";

/*****************************
     Globals / All globals have static duration
*****************************/
// Time variables
static char GMT_string[30]               = "Tue, 15 Nov 1994 08:12:31 GMT"; /*!< GMT String used in header response */
static const code char* days_in_week[]   = {"Mon","Tue","Wed","Thu","Fri","Sat","Sun"}; /*!< 3 char days used in GMT string */
static const code char* months_in_year[] = {"Nul","Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};/*!< 3 char months starts with 1..12 */
static volatile uint32_t ntp_epoch_time  = 0; /*!< UNIX timestamp returned from NTP server, updated by ISR */

// Server variables
static webpage_t pages[MAX_PAGES];
static int8_t page_number                = 0;

static uint16_t http_port                = 0;      /*!< Port that HTTP server serves pages */
static char get_post_buffer[GETPOSTSIZE];

// Server modes
static bool irqModeEnable                = FALSE;  /*!<  */
static volatile bool packet_flag         = FALSE;  /*!<  */

// Form variables
static key_value_t form_fields[MAX_FIELDS];        /*!<  */
// Downloaded File variable
static char downloaded_filename[FILENAMESIZE];

/*!< List of server elements and sockets */
static uint8_t stackTop                  = -1;     /*!<  */
static socket_t sockets[MAX_SOCKET], *socketStack[MAX_SOCKET], *socketInUse[MAX_SOCKET];              /*!<  */

// Function pointers to external definitions
void ( *form_received )( key_value_t* values );    /*!<  */
void ( *file_download )( char* fileName );         /*!<  */
void ( *ntp_received  )( uint32_t time );          /*!<  */

/*****************************
     Helper Function Proto
*****************************/
// String functions
static void strcpy_const( char* destination, const char* source );
static void strcat_const( char* destination, const char* source );
static int strncmp_const( char* str1, const char* str2, int num );
//static uint32_t strlen_const ( const char* str );
static void update_GMT_str( char* gmt_str );

#if AVR
// Additional AVR functions not included with library
static void Net_Ethernet_28j60_Reset( void );
static uint16_t Net_Ethernet_28j60_bufferFreeSize( SOCKET_28j60_Dsc* socket_Dsc );
#endif

// General socket functions
static int  stack_push          ( socket_t** stack, socket_t* socket );
static socket_t* stack_pop      ( socket_t** stack );
// Functions used in server requests
static uint8_t get_chunked_size ( uint16_t* new_size );

/*************************************
 * Server / Client socket functions
 ************************************/
// Step 0
static socket_t* init_new_socket( SOCKET_Dsc* eth_socket );
// Step 1
static void get_client_request  ( socket_t* const socket );
// Step 2
static void parse_client_request( socket_t* const socket );
// Step 3
static void process_get_request ( socket_t* const socket );
// Step 4
static void find_requested_file ( socket_t* const socket );
// Step 5
static void serve_headers       ( socket_t* const socket );
// Step 6
static void serve_body          ( socket_t* const socket );
// Step 7

// Step 8
static void close_socket        ( socket_t* const socket );
/************************************
 *     Server Request functions
 ***********************************/
// Step 1
static void server_send_request ( socket_t* const socket );
// Step 2
static void get_server_response ( socket_t* const socket );
// Step 3
static void parse_server_response( socket_t* const socket );
// Step 4
static void open_server_file    ( socket_t* const socket );
// Step 5
static void get_server_data     ( socket_t* const socket );
// Step 6
static void close_server_socket ( socket_t* const socket );


/*************************************
 *    Functions
 ************************************/
#if AVR
extern sbit Net_Ethernet_28j60_Rst;
extern sbit Net_Ethernet_28j60_CS;
extern sbit Net_Ethernet_28j60_Rst_Direction;
extern sbit Net_Ethernet_28j60_CS_Direction;
extern sfr sbit Net_Ethernet_28j60_INT;
extern sfr sbit Net_Ethernet_28j60_INT_Direction;

static void Net_Ethernet_28j60_Reset()
{
    Net_Ethernet_28j60_CS = 0;
    SPI1_Write( 0xFF );
    Net_Ethernet_28j60_CS = 1;
}

static uint16_t Net_Ethernet_28j60_bufferFreeSize( SOCKET_28j60_Dsc* socket_Dsc )
{
    return ( TCP_TX_SIZE_28j60 - socket_Dsc->lastWritten );
}
#endif
/*************************************
 *    Utility Functions
 ************************************/
/*!
 *  \brief Copies a ROM string to RAM
 *
 *  \note
 *   <notes>
 */
static void strcpy_const( char* destination, const char* source )
{
    while( *destination++ = *source++ );

    *destination = '\0';
}
/*!
 *  \brief Concatenates a ROM string to RAM
 *
 *  \note
 *   <notes>
 */
static void strcat_const( char* destination, const char* source )
{
    while( *destination != '\0' )
    {
        destination++;
    }

    strcpy_const( destination, source );
}
/*!
 *  \brief Compares a ROM string to RAM
 *
 *  \note
 *   <notes>
 */
static int strncmp_const( char* str1, const char* str2, int num )
{
    while( num-- != 0 )
    {
        if( *str1++ != *str2++ )
        {
            return 1;
        }

    }

    return 0;
}
/*static uint32_t strlen_const ( const char* str )
{
    uint32_t str_size = 0;

    while( *str != '\0' && str_size < INT32_MAX )
    {
        str++;
        str_size++;
    }

    return str_size;
}*/

/*
 *  Updates the GMT time string
 */
static void update_GMT_str( char* gmt_str )
{
    //Tue, 15 Nov 1994 08:12:31 GMT
    TimeStruct ts1;

    Time_epochToDate( ntp_epoch_time, &ts1 );

    gmt_str[0] = days_in_week[ts1.wd][0];
    gmt_str[1] = days_in_week[ts1.wd][1];
    gmt_str[2] = days_in_week[ts1.wd][2];
    gmt_str[3] = ',';
    gmt_str[4] = ' ';
    gmt_str[5] = ( ts1.md / 10 ) + 48;
    gmt_str[6] = ( ts1.md % 10 ) + 48;
    gmt_str[7] = ' ';
    gmt_str[8]  = months_in_year[ts1.mo][0];
    gmt_str[9]  = months_in_year[ts1.mo][1];
    gmt_str[10] = months_in_year[ts1.mo][2];
    gmt_str[11] = ' ';
    gmt_str[12] = '2';
    gmt_str[13] = '0';
    gmt_str[14] = ( ( ts1.yy - 2000 ) / 10 ) + 48;
    gmt_str[15] = ( ( ts1.yy - 2000 ) % 10 ) + 48;
    gmt_str[16] = ' ';
    gmt_str[17] = ( ts1.hh / 10 ) + 48;
    gmt_str[18] = ( ts1.hh % 10 ) + 48;
    gmt_str[19] = ':';
    gmt_str[20] = ( ts1.mn / 10 ) + 48;
    gmt_str[21] = ( ts1.mn % 10 ) + 48;
    gmt_str[22] = ':';
    gmt_str[23] = ( ts1.ss / 10 ) + 48;
    gmt_str[24] = ( ts1.ss % 10 ) + 48;
    gmt_str[25] = ' ';
    gmt_str[26] = 'G';
    gmt_str[27] = 'M';
    gmt_str[28] = 'T';
    gmt_str[29] = '\0';
}

/*************************************
 *    Stack Functions
 ************************************/
/*!
 *  \brief Pushes socket onto stack
 *
 */
static int stack_push( socket_t** stack, socket_t* socket )
{
    int socket_index;

    if( stackTop == ( MAX_SOCKET-1 ) )
    {
        return -1;
    }

    else
    {
        // Reset the socket back to 0
        memset( socket->buffer_large, 0, LARGEBUFFERSIZE );
        memset( socket->file_name,    0, FILENAMESIZE );
        socket->buffer_position = 0;
        socket->socketSM   = SM_HTTP_IDLE;
        socket->webpage    = NULL;
        socket->eth_socket = NULL;

        stack[++stackTop] = socket;

        for( socket_index = 0; socket_index < MAX_SOCKET; socket_index++ )
        {
            if( socketInUse[socket_index] == socket )
            {
                socketInUse[socket_index] = NULL;
                break;
            }
        }

        return 0;
    }

}

/*!
 *  \brief Removes socket from stack
 *
 */
static socket_t* stack_pop( socket_t** stack )
{
    int i;

    if ( stackTop == -1 )
    {
        return NULL;
    }

    else
    {
        socket_t* temp = stack[stackTop];
        stack[stackTop--] = 0;

        for( i = 0; i < MAX_SOCKET; i++ )
        {
            if( socketInUse[i] == NULL )
            {
                socketInUse[i] = temp;
                break;
            }
        }

        return temp;
    }
}

/*************************************
 *    Gen Socket Functions
 ************************************/
/*!
 *  \brief Initializes new sockets
 *
 */
static socket_t* init_new_socket( SOCKET_Dsc* eth_socket )
{
    int socketCount;
    socket_t* newSocket = NULL;

    for( socketCount = 0; socketCount < MAX_SOCKET; socketCount++ )
    {
        if( socketInUse[socketCount] != NULL )
        {
            if( socketInUse[socketCount]->eth_socket == eth_socket )
            {
                return socketInUse[socketCount];
            }
        }
    }

    newSocket = stack_pop( socketStack );
    // Assign socket
    newSocket->eth_socket = eth_socket;
    newSocket->socketSM = SM_HTTP_GET_REQUEST;

    return newSocket;
}

/*****************************
 *  Server Request Functions
 ****************************/
/*!
 *  \brief Gets size of transfer encoding chunk
 *
 *  \param address of int that will be populated with result
 *
 *  \return
 *     \retval int: number of bytes removed from rx buffer
 *
 *  \note
 *   <notes>
 */
static int get_chunked_size( uint16_t* new_size )
{
    /* Chunked encoding keeps the length of the response in following hex chars followed by a CR and LF */
    char length_str[5] = {0};
    char* p_length = length_str;
    int bytes_read = 0;
    char tmp_buffer;
    bool terminate = FALSE;

    while( terminate != TRUE && bytes_read++ < 8 )
    {
        tmp_buffer = Net_Ethernet_getByte();

        switch( tmp_buffer )
        {
            case CR:  // Get rid of those nasty carriage returns
                break;

            case LF:  // Ah, found a line feed.
                if( bytes_read > 2 )
                {
                    terminate = TRUE;
                    *new_size = xtoi( length_str );
                }

                break;

            default:
                *p_length++ = tmp_buffer;
        }
    }

    return bytes_read;
}

/*******************************
 * Client TCP handler functions
 ******************************/
/*!
 *  \brief Step 1 - Get Client Request
 *
 *  \param <Parameters Accepted>
 *
 *  \note
 *   <notes>
 */
static void get_client_request( socket_t* const socket )
{
    uint16_t len = socket->eth_socket->dataLength;

    if( len == 0 )
    {
        close_socket( socket );
    }

    else
    {
        char window_buffer[MY_MSS_SIZE];

        if( LARGEBUFFERSIZE < ( len + socket->buffer_position ) )
        {
            Net_Ethernet_getBytes( window_buffer, 0xFFFF, len );
            strcat( socket->buffer_large, window_buffer );
            socket->buffer_position += len;
            
            if( len < MY_MSS_SIZE )
            {
                socket->buffer_position = 0;
                socket->socketSM = SM_HTTP_PARSE_REQUEST;
            }
        }
        else
        {
            while( socket->buffer_position < LARGEBUFFERSIZE )
            {
                socket->buffer_large[socket->buffer_position++] = Net_Ethernet_getByte();
            }
            socket->buffer_position = 0;
            socket->socketSM = SM_HTTP_PARSE_REQUEST;
        }
    }
}

/*!
 *  \brief Step 2 - Parse the client request
 *
 *  \param <Parameters Accepted>
 *
 *  \note
 *   <notes>
 */
static void parse_client_request( socket_t* const socket )
{
    char* p_parse;
    char* const p_buffer = socket->buffer_large; // Alias that saves a little typing

    // This is put in place in case we don't have a valid GET or POST on the request
    socket->socketSM = SM_HTTP_DISCONNECT;

    // Tokenize the entire huge string by \n which
    p_parse = strtok( p_buffer, "\n" );

    while( p_parse != NULL )
    {
        if( strncmp( p_parse, "GET ", 4 ) == 0 )
        {
            socket->socketSM = SM_HTTP_PROCESS_GET;
            strcpy( get_post_buffer, p_parse );
        }

        else if( strncmp( p_parse, "POST ", 4 ) == 0 )
        {
            socket->socketSM = SM_HTTP_PROCESS_POST;
            strcpy( get_post_buffer, p_parse );
        }

        else if( strncmp( p_parse, "Conne", 4 ) == 0 )
        {
            char* p_connect;
            p_connect = strchr( p_parse, ' ' );
            p_connect++;

            if( strncmp( p_connect, "keep-alive", 2 ) == 0 )
            {
                socket->connect_type = CONNECTION_KEEP_ALIVE;
            }

            else if( strncmp( p_connect, "close", 2 ) == 0 )
            {
                socket->connect_type = CONNECTION_CLOSE;
            }
        }

        /*else if( strncmp( p_parse, "Content-Length: ", 10 ) == 0 )
        {
            //TODO: This is for POST requests and file length attached
        }

        else if( strncmp( p_parse, "Host:", 4 ) == 0 )
        {
          //TODO: Needs to be present for http 1.1 compliance

        }*/

        p_parse = strtok( NULL, "\n" );
    }
}

/*!
 *  \brief Step 3 - Process GET request
 *
 *  \param <Parameters Accepted>
 *
 *  \note
 *   <notes>
 */
static void process_get_request( socket_t* const socket )
{
    char* p_buffer = get_post_buffer;
    char* p_parser;

    // Past the GET and now at the request
    while( *p_buffer++ != ' ' );

    // Let's look at the string and see if this has any form data
    if( strchr( p_buffer, '?' ) == NULL )
    {
        // No form data. How about a file name
        if( strncmp( p_buffer, "/ ", 2 ) == NULL )
        {
            strcpy( socket->file_name, "index.htm" );
            socket->socketSM = SM_HTTP_FIND_FILE;
        }

        else
        {
            char* tmp_parser;
            int i = 0, j = 0;

            p_parser = strchr( p_buffer, ' ' );
            // Get rid of the HTTP/1.1 at the end by inserting a null terminator
            *p_parser = '\0';

            p_parser = strchr( p_buffer, '/' );

            while( p_parser != NULL )
            {
                p_parser++;
                tmp_parser = p_parser;
                p_parser = strchr( p_parser, '/' );
            }

            while( ( *tmp_parser != '.' ) && ( i < 8 ) )
            {
                socket->file_name[i++] = *tmp_parser++;
            }

            tmp_parser++;
            socket->file_name[i++] = '.';

            while( ( *tmp_parser != '\0' && j < 3 ) )
            {
                socket->file_name[i++] = *tmp_parser++;
                j++;
            }

            socket->file_name[i] = '\0';

            socket->socketSM = SM_HTTP_FIND_FILE;
        }
    }

    else
    {
        int i = 0, j = 0;
        //What is given to this is something that looks like: /tut.htm?check0=on&text0=Richard&check1=on&text1=Lowe&check2=on&text2=ocdrichard%40gmail.com&check3=on&text3=05%2F25%2F1974
        p_buffer++;  // gets us past the /

        while( *p_buffer != '.' )
        {
            socket->file_name[i++] = *p_buffer++;
        }

        p_buffer++;
        socket->file_name[i++] = '.';

        while( ( *p_buffer != '?' && j < 3 ) )
        {
            socket->file_name[i++] = *p_buffer++;
            j++;
        }

        // That takes care of the file name, now the form values

        socket->file_name[i] = '\0';
        p_buffer++;

        i = 0;
        j = 0;

        p_parser = strchr( p_buffer, ' ' );
        *p_parser = '\0';
        p_parser = strtok( p_buffer, "&" );

        while( p_parser != NULL && i < MAX_FIELDS )
        {
            j = 0;

            while( *p_parser != '=' && j < MAX_FIELDS_KEY )
            {
                form_fields[i].key[j++] = *p_parser++;
            }

            form_fields[i].key[j] = '\0';

            p_parser++;
            j = 0;

            while( *p_parser != '\0' && j < MAX_FIELDS_VALUE )
            {
                form_fields[i].value[j++] = *p_parser++;
            }

            form_fields[i].value[j] = '\0';

            i++;

            form_fields[i].key[0] = '\0';
            form_fields[i].value[0] = '\0';

            p_parser = strtok( NULL, "&" );
        }

        socket->socketSM = SM_HTTP_FIND_FILE;

        form_received( form_fields );
    }
}

/*!
 *  \brief Step 3 - Process POST request ( NOT YET IMPLEMENTED )
 *
 *  \pre <Preconditions that need to be met before called>
 *
 *  \param <Parameters Accepted>
 *
 *  \note
 *   <notes>
 */
static void process_post_request( socket_t* const socket )
{


}

/*!
 *  \brief Step 4 - Validate file
 *
 *  \pre <Preconditions that need to be met before called>
 *
 *  \param <Parameters Accepted>
 *
 *  \note
 *   <notes>
 */
static void find_requested_file( socket_t* const socket )
{
    int page_index;

    socket->socketSM = SM_HTTP_FILE_NOT_FOUND;

    for( page_index = 0; page_index < page_number; page_index++ )
    {
        // TODO: May need to change to strcmp instead of memcmp
        if( strcmp( socket->file_name, pages[page_index].page_name ) == 0 )
        {
            // Assign a webpage
            socket->webpage  = &pages[page_index];
            // Ready for generation of header
            socket->socketSM = SM_HTTP_SERVE_HEADERS;

            if( socket->webpage->static_page_flag == FALSE )
            {
                char txt_buffer[12];
                uint32_t p_size;
                // Callback that updates data
                pages[page_index].dynamic_page.page_update();
                // Refresh size of dynamic page
                p_size = strlen( pages[page_index].dynamic_page.start_ptr );
                // Convert size to string
                LongToStr( p_size, txt_buffer );
                strcpy( pages[page_index].page_size_str, Ltrim( txt_buffer ) );
                // Refresh dynamic page end pointer
                pages[page_index].dynamic_page.end_ptr = pages[page_index].dynamic_page.start_ptr + p_size;

            }

            break;
        }
    }
}

/*!
 *  \brief Step 5 - Serve headers
 *
 *  \pre <Preconditions that need to be met before called>
 *
 *  \param <Parameters Accepted>
 *
 *  \note
 *   <notes>
 */
static void serve_headers( socket_t* const socket )
{
    // Header
    Net_Ethernet_putConstStringTCP( http_response_header,           socket->eth_socket );
    // Timestamp in GMT format
    Net_Ethernet_putStringTCP     ( GMT_string,                     socket->eth_socket );
    // Length
    Net_Ethernet_putConstStringTCP( http_response_length,           socket->eth_socket );
    Net_Ethernet_putStringTCP     ( socket->webpage->page_size_str, socket->eth_socket );
    // Content type
    Net_Ethernet_putConstStringTCP( http_response_type,             socket->eth_socket );
    Net_Ethernet_putConstStringTCP( http_content_types[socket->webpage->page_type], socket->eth_socket );

    socket->socketSM = SM_HTTP_SERVE_BODY;
}

/*!
 *  \brief Step 6 - Serve body
 *
 *  \param <Parameters Accepted>
 *
 *  \note
 *   <notes>
 */
static void serve_body( socket_t* const socket )
{
    bool done_flag = FALSE;              // Flag that represents done sending body
    uint16_t remaining_bytes;            // Bytes remaining from current to end

    // Loop for static pages located in ROM
    if( socket->webpage->static_page_flag == TRUE )
    {
        // Get remaining bytes
        remaining_bytes = ( socket->webpage->static_page.end_ptr - socket->webpage->static_page.current_ptr );
        // Increments current pointer by how many bytes were successfully written to buffer
        socket->webpage->static_page.current_ptr += Net_Ethernet_putConstBytesTCP( socket->webpage->static_page.current_ptr,
                remaining_bytes,
                socket->eth_socket );

        // If current pointer == end pointer, transmit complete
        if( socket->webpage->static_page.current_ptr == socket->webpage->static_page.end_ptr )
        {
            done_flag = TRUE;
            socket->webpage->static_page.current_ptr = socket->webpage->static_page.start_ptr;
        }
    }

    // Loop for dynamic pages located RAM
    else
    {
        // Get remaining bytes
        remaining_bytes = ( socket->webpage->dynamic_page.end_ptr - socket->webpage->dynamic_page.current_ptr );
        // Increments current pointer by how many bytes were successfully written to buffer
        socket->webpage->dynamic_page.current_ptr += Net_Ethernet_putBytesTCP( socket->webpage->dynamic_page.current_ptr,
                remaining_bytes,
                socket->eth_socket );

        // If current pointer == end pointer, transmit complete
        if( socket->webpage->dynamic_page.current_ptr == socket->webpage->dynamic_page.end_ptr )
        {
            done_flag = TRUE;
            socket->webpage->dynamic_page.current_ptr = socket->webpage->dynamic_page.start_ptr;
        }
    }

    // Checks the flag, if TRUE, transmission of body complete
    if( done_flag == TRUE )
    {
        switch( socket->connect_type )
        {
            case CONNECTION_CLOSE:
                socket->socketSM = SM_HTTP_WAIT_TILL_DONE;
                break;

            case CONNECTION_KEEP_ALIVE:
                memset( socket->buffer_large, 0, LARGEBUFFERSIZE );
                memset( socket->file_name,    0, FILENAMESIZE );
                socket->webpage  = NULL;
                socket->socketSM = SM_HTTP_GET_REQUEST;
                break;
        }
    }
}

/*!
 *  \brief Step 8 - Close connection
 *
 *  \param <Parameters Accepted>
 *
 *  \note
 *   <notes>
 */
static void close_socket( socket_t* const socket )
{
    // If socket happens to be download socket
    if( socket->eth_socket == server_request.eth_socket_download )
    {
        close_server_socket( socket );
    }

    else
    {
        if( Net_Ethernet_disconnectTCP( socket->eth_socket ) == 1 )
        {
            // If the page was found, it has a page assigned, if not it would be a null reference
            if( socket->webpage != NULL )
            {
                // Is it a static or dynamic page
                if( socket->webpage->static_page_flag == TRUE )
                {
                    // Reset the current pointer to start
                    socket->webpage->static_page.current_ptr = socket->webpage->static_page.start_ptr;
                }

                else
                {
                    // Reset the current pointer to start
                    socket->webpage->dynamic_page.current_ptr = socket->webpage->dynamic_page.start_ptr;
                }
            }

            // Push it back onto the stack of available sockets
            stack_push( socketStack, socket );
        }

        else
        {
            debug_print( "Could not close socket\r\n" );
        }
    }
}



/************************************
 *        Server Request Functions
 ***********************************/
/*!
 *  \brief SERVER Step 1
 *
 *  \param <Parameters Accepted>
 *
 *  \note
 *   <notes>
 */
static void server_send_request( socket_t* const socket )
{
    // Step 1. The request needs to be sent to the server
    Net_Ethernet_putBytesTCP( socket->buffer_large, server_request.transmit_size, socket->eth_socket );
    //memset( socket->buffer_large, 0, LARGEBUFFERSIZE );
    socket->socketSM = SM_HTTP_SVR_GET_RESPONSE;
}

/*!
 *  \brief SERVER Step 2
 *
 *  \pre <Preconditions that need to be met before called>
 *
 *  \param <Parameters Accepted>
 *
 *  \note
 *   <notes>
 */
static void get_server_response( socket_t* const socket )
{
    uint16_t len         = socket->eth_socket->dataLength;
    uint8_t termination  = 0;
    char buffer;

    while( len-- > 0 )
    {
        buffer = Net_Ethernet_getByte();

        switch( buffer )
        {
            case CR:  // Get rid of those nasty carriage returns
                break;

            case LF:  // Ah, found a line feed.
                socket->buffer_large[server_request.position++] = buffer;
                termination++;

                // If \n\n is found, then terminate the loop and parse
                if( termination == 2 )
                {
                    server_request.remaining = len;
                    socket->buffer_large[server_request.position] = '\0';
                    socket->socketSM = SM_HTTP_SVR_PARSE_RESPONSE;
                    return;
                }

                break;

            default:
                socket->buffer_large[server_request.position++] = buffer;
                termination = 0;
        }

        if( server_request.position == LARGEBUFFERSIZE )
        {
            server_request.remaining = len;
            socket->buffer_large[server_request.position] = '\0';
            socket->socketSM = SM_HTTP_SVR_PARSE_RESPONSE;
            return;
        }
    }
}

/*!
 *  \brief SERVER Step 3
 *
 *  \pre <Preconditions that need to be met before called>
 *
 *  \param <Parameters Accepted>
 *
 *  \note
 *   <notes>
 */
static void parse_server_response( socket_t* const socket )
{
    char* p_parser;
    char* const p_buffer = socket->buffer_large;

    p_parser = strtok( p_buffer, "\n" );

    if( p_parser == NULL )
    {
        debug_print( "Not valid header\r\n" );
        socket->socketSM = SM_HTTP_SVR_DISCONNECT;
        return;
    }

    else if( ( strncmp_const( p_parser, http_response_OK , 13 ) == 0 )
             || ( strncmp_const( p_parser, http_response_OK_old , 13 ) == 0 ) )
    {
        p_parser = strtok( NULL, "\n" );

        while( p_parser != NULL )
        {
            //debug_print( p_parser );
            //debug_print( "\r\n" );

            if( p_parser[0] == 'C' )
            {
                switch( p_parser[8] )
                {
                    case 'D':
                    {
                        int i = 0;
                        int j = 0;
                        char* ptr;

                        ptr = strchr( p_parser, '"' );
                        ptr++;

                        while( *( ptr ) != '.' && i < 8 )
                        {
                            socket->file_name[i++] = *ptr++;
                        }

                        if( *ptr != '.' )
                        {
                            while( *( ++ptr ) != '.' );
                        }

                        socket->file_name[i++] = *ptr++;

                        while( *( ptr ) != '"' && j < 3 )
                        {
                            socket->file_name[i++] = *( ptr++ );
                            j++;
                        }

                        socket->file_name[i] = '\0';
                        strcpy( downloaded_filename, socket->file_name );

                        break;
                    }

                    case 'L':
                        server_request.file_length = atol( strchr( p_parser, ' ' ) );

                        socket->connect_type = CONNECTION_KEEP_ALIVE;

                        if( server_request.file_length == 0 )
                        {
                            socket->socketSM = SM_HTTP_SVR_DISCONNECT;
                            return;
                        }

                        break;

                    case 'T':
                        //strcpy( header.contentType, ptr );
                        break;

                    case 'o':
                    {
                        char* p_connect;
                        p_connect = strchr( p_parser, ' ' );

                        p_connect++;

                        if( strncmp( p_connect, "keep-alive", 6 ) == 0 )
                        {
                            socket->connect_type = CONNECTION_KEEP_ALIVE;
                        }

                        else if( strncmp( p_connect, "close", 5 ) == 0 )
                        {
                            socket->connect_type = CONNECTION_CLOSE;
                        }

                        break;
                    }
                }
            }

            else if( p_parser[0] == 'T' )
            {
                if( p_parser[9] == 'E' )
                {
                    char* p_encode;
                    p_encode = strchr( p_parser, ' ' );
                    p_encode++;

                    if( strncmp( p_encode, "chunked", 3 ) == 0 )
                    {
                        server_request.position = 0;
                        socket->encoding_type   = ENCODING_CHUNKED;
                    }
                }
            }

            p_parser = strtok( NULL, "\n" );
        }

        socket->socketSM = SM_HTTP_SVR_OPEN_FILE;
    }

    else
    {
        debug_print( "Not a valid response\r\n" );
        socket->socketSM = SM_HTTP_SVR_DISCONNECT;
    }
}

/*!
 *  \brief SERVER Step 4
 *
 *  \pre <Preconditions that need to be met before called>
 *
 *  \param <Parameters Accepted>
 *
 *  \note
 *   <notes>
 */
static void open_server_file( socket_t* const socket )
{
    // Swapfile creation and opening
    if( FAT32_Exists( SWAPFILE ) == 1 )
    {
        FAT32_Delete( SWAPFILE );
    }

    if( FAT32_MakeSwap( SWAPFILE, 1000, &server_request.swap_sector ) == 0 )
    {
        server_request.swap_handle = FAT32_Open( SWAPFILE, FILE_WRITE );

        if( server_request.swap_handle < 0 )
        {
            debug_print( "Could not open swapfile\r\n" );
            socket->socketSM = SM_HTTP_SVR_DISCONNECT;
        }

        else
        {
            /* Hold length of packet datalength, this SHOULD be less than MY_MSS_SIZE */
            uint16_t len = server_request.remaining;
            char buffer[MY_MSS_SIZE] = {0};

            /* Chunked encoding implimentation */
            if( socket->encoding_type == ENCODING_CHUNKED )
            {
                int i;
                char* p_buffer = buffer;

                for( i = 0; i < len; i++ )
                {
                    if( server_request.position == 0 )
                    {
                        i += get_chunked_size( &server_request.position );
                    }

                    *p_buffer++ = Net_Ethernet_getByte();
                    server_request.position--;
                }

                len = strlen( buffer );
            }

            else
            {
                // Retrieve the entire buffer remaining from the first read
                Net_Ethernet_getBytes( buffer, 0xFFFF, len );  /*<! Get entire window and copy to buffer */
            }

            // Commit to swapfile
            if( FAT32_WriteSwap( server_request.swap_handle, buffer, len ) != 0 )
            {
                debug_print( "Error in writting to swap\r\n" );
                socket->socketSM = SM_HTTP_SVR_DISCONNECT;
            }

            else
            {
                // Must add to capture length total
                server_request.capture_length += len;
                socket->socketSM = SM_HTTP_SVR_GET_DATA;
            }
        }
    }

    else
    {
        debug_print( "Could not create swapfile\r\n" );
        socket->socketSM = SM_HTTP_SVR_DISCONNECT;
    }
}

/*!
 *  \brief SERVER Step 5
 *
 *  \param <Parameters Accepted>
 *
 *  \note
 *   <notes>
 */
static void get_server_data( socket_t* const socket )
{
    char buffer[MY_MSS_SIZE] = {0};
    uint16_t len = socket->eth_socket->dataLength; // Hold length of packet datalength
    bool download_flag = FALSE;

    // Chunked encoding implimentation
    if( socket->encoding_type == ENCODING_CHUNKED )
    {
        int i;
        char* p_buffer = buffer;

        for( i = 0; i < len; i++ )
        {
            if( server_request.position == 0 )
            {
                i += get_chunked_size( &server_request.position );

                if( server_request.position == 0 )
                {
                    download_flag = TRUE;
                    break;
                }
            }

            *p_buffer++ = Net_Ethernet_getByte();
            server_request.position--;
        }

        len = strlen( buffer );
    }

    else
    {
        Net_Ethernet_getBytes( buffer, 0xFFFF, len );  /*<! Get entire window and copy to buffer */
    }

    // Write all the data to the swap
    if( FAT32_WriteSwap( server_request.swap_handle, buffer, len ) != 0 )
    {
        debug_print( "Error in writting to swap\r\n" );
        socket->socketSM = SM_HTTP_SVR_DISCONNECT;
    }

    else
    {
        server_request.capture_length += len;        /*<! Increase capture count */

        // Check for the byte count to reach length size
        switch ( socket->connect_type )
        {
            case CONNECTION_KEEP_ALIVE:
                if( server_request.capture_length >= server_request.file_length )
                {
                    debug_print( "\nKeep alive - Done!\r\n" );
                    socket->socketSM = SM_HTTP_SVR_DISCONNECT;
                }

                break;

            case CONNECTION_CLOSE:
                if( server_request.file_length == 0 )
                {
                    if( download_flag == TRUE )
                    {
                        debug_print( "\nClose - DONE!\r\n" );
                        socket->socketSM = SM_HTTP_SVR_DISCONNECT;
                    }
                }

                else
                {
                    if( server_request.capture_length >= server_request.file_length )
                    {
                        debug_print( "\nClose: Length included - Done!\r\n" );
                        socket->socketSM = SM_HTTP_SVR_DISCONNECT;
                    }
                }

                break;
        }
    }
}

/*!
 *  \brief Close server socket
 *
 *  \param <Parameters Accepted>
 *
 *  \note
 *   <notes>
 */
static void close_server_socket( socket_t* const socket )
{
    if( Net_Ethernet_disconnectTCP( socket->eth_socket ) == 1 )
    {
        bool file_ready = FALSE;

        // To keep filesystem integrity we need to do a good job of making sure we don't leave files open
        if( FAT32_GetFileHandle( SWAPFILE, &server_request.swap_handle ) != -1 )
        {
            // Issue close if open
            FAT32_Close( server_request.swap_handle );
        }

        // Represents we have bytes on the buffer to write
        if( server_request.capture_length > 0 )
        {
            if( FAT32_Exists( SWAPFILE ) == 1 )
            {
                server_request.swap_handle = FAT32_Open( SWAPFILE, FILE_READ );     // Open swap for reading

                if( FAT32_Exists( socket->file_name ) == 1 )
                {
                    FAT32_Delete( socket->file_name );
                }

                server_request.download_handle = FAT32_Open( socket->file_name, FILE_WRITE );

                if( ( server_request.swap_handle < 0 ) || ( server_request.download_handle < 0 ) )
                {
                    FAT32_Close( server_request.swap_handle );
                    FAT32_Close( server_request.download_handle );
                    FAT32_Delete( SWAPFILE );
                }

                else
                {
                    char txt_buf[50];
                    // Read and write 1 sector at a time, this is the buffer to hold it
                    char p_data_buffer[SECTOR_SIZE];
                    uint16_t sectors = ( server_request.capture_length / SECTOR_SIZE ); // How many sectors to write
                    uint16_t remain  = ( server_request.capture_length % SECTOR_SIZE ); // How many remaining bytes

                    sprinti( txt_buf, "Position is still: %u\r\n", server_request.position );
                    debug_print( txt_buf );

                    while( sectors-- != 0 )
                    {
                        FAT32_ReadSwap( server_request.swap_handle,     p_data_buffer, SECTOR_SIZE );
                        FAT32_Write   ( server_request.download_handle, p_data_buffer, SECTOR_SIZE );
                    }

                    // These are the remaining bytes to read and write
                    FAT32_ReadSwap( server_request.swap_handle,     p_data_buffer, remain );
                    FAT32_Write   ( server_request.download_handle, p_data_buffer, remain );
                    // CRITICAL!!! CLOSE FILE HANDLES
                    FAT32_Close ( server_request.swap_handle );
                    FAT32_Close ( server_request.download_handle );
                    // Make sure we delete swapfile
                    FAT32_Delete( SWAPFILE );
                    // Copy our new downloaded file name to the global
                    strcpy( downloaded_filename, socket->file_name );

                    file_ready = TRUE;
                }
            }
        }

        // Reset the download socket to NULL ready for another
        server_request.eth_socket_download = NULL;
        server_request.capture_length = 0;
        server_request.file_length    = 0;
        server_request.position       = 0;
        stack_push( socketStack, socket );

        if( file_ready == TRUE )
        {
            file_download( downloaded_filename );
        }
    }
}

/************************************
 *
 *     Handlers
 *
 ************************************/
/*!
 *  \brief Handles all TCP packets
 *
 *  \param SOCKET_28j60_Dsc *socket - socket calling from do packet
 *
 */
void Net_Ethernet_UserTCP( SOCKET_Dsc* eth_socket )
{
    if( eth_socket->destPort == http_port )
    {
        socket_t* current_socket = init_new_socket( eth_socket );

        if( current_socket != NULL )
        {
            // Reset timer to keep socket alive
            current_socket->alive_time = 0;

            /* Each socket has a state machine attached.  This state determines what stage of service it is in.*/
            switch( current_socket->socketSM )
            {
                case SM_HTTP_GET_REQUEST:
                    // Step 1. Get request from client
                    get_client_request( current_socket );

                    if( current_socket->socketSM != SM_HTTP_PARSE_REQUEST )
                    {
                        break;
                    }

                case SM_HTTP_PARSE_REQUEST:
                    // Step 2. Parse incoming request, AVR 2ms
                    parse_client_request( current_socket );

                    if( current_socket->socketSM != SM_HTTP_PROCESS_GET )
                    {
                        break;
                    }

                case SM_HTTP_PROCESS_GET:
                    // Step 3. process
                    process_get_request( current_socket );

                    if( current_socket->socketSM != SM_HTTP_FIND_FILE )
                    {
                        break;
                    }

                case SM_HTTP_FIND_FILE:
                    // Step 4. Find file, AVR 43ms
                    find_requested_file( current_socket );

                    /*validate_file( current_socket );*/
                    if( current_socket->socketSM == SM_HTTP_FILE_NOT_FOUND )
                    {
                        Net_Ethernet_putConstStringTCP( http_response_not_found, current_socket->eth_socket );

                        if( current_socket->connect_type == CONNECTION_CLOSE )
                        {
                            current_socket->socketSM = SM_HTTP_DISCONNECT;
                        }

                        else
                        {
                            memset( current_socket->file_name, 0, FILENAMESIZE );
                            memset( current_socket->buffer_large, 0, LARGEBUFFERSIZE );
                            current_socket->socketSM = SM_HTTP_GET_REQUEST;
                        }

                        break;
                    }

                case SM_HTTP_SERVE_HEADERS:
                    // Step 5. Serve Header
                    serve_headers( current_socket );

                case SM_HTTP_SERVE_BODY:
                    // Step 6. Serve body
                    serve_body( current_socket );

                    if( current_socket->socketSM != SM_HTTP_WAIT_TILL_DONE )
                    {
                        break;
                    }

                case SM_HTTP_WAIT_TILL_DONE:

                    // Step 7. Wait until buffer empty
                    if( Net_Ethernet_bufferEmptyTCP( current_socket->eth_socket ) != 1 )
                    {
                        break;
                    }

                    current_socket->socketSM = SM_HTTP_DISCONNECT;

                case SM_HTTP_DISCONNECT:
                    // Step 8. Disconnect and remove from list
                    close_socket( current_socket );
                    break;
            }
        }
    }

    else if( eth_socket->destPort == REQUESTPORT )
    {
        socket_t* current_socket = init_new_socket( eth_socket );

        // Reset timer to keep socket alive
        current_socket->alive_time = 0;

        switch( current_socket->socketSM )
        {
            case SM_HTTP_SVR_SEND_REQUEST:
                // Step 1. Send request to server
                server_send_request( current_socket );
                break;

            case SM_HTTP_SVR_GET_RESPONSE:
                // Step 2. Server is responding.  Get whole request
                get_server_response( current_socket );

                if( current_socket->socketSM != SM_HTTP_SVR_PARSE_RESPONSE )
                {
                    break;
                }

            case SM_HTTP_SVR_PARSE_RESPONSE:
                // Step 3. Parse the response
                parse_server_response( current_socket );

                if( current_socket->socketSM != SM_HTTP_SVR_OPEN_FILE )
                {
                    debug_print( "Bad things happened\r\n" );
                    close_socket( current_socket );
                    break;
                }

            case SM_HTTP_SVR_OPEN_FILE:
                // Step 4. Open the file and get what bytes remain on the buffer
                open_server_file( current_socket );
                break;

            case SM_HTTP_SVR_GET_DATA:
                // Step 5. Get all that sweet data
                get_server_data( current_socket );

                if( current_socket->socketSM != SM_HTTP_SVR_DISCONNECT )
                {
                    break;
                }

            case SM_HTTP_SVR_DISCONNECT:
                // Step 6. Close it down
                close_server_socket( current_socket );
                break;
        }
    }
}

/*!
 *  \brief Handles all UDP packets
 *
 *  \param Struct of UDP socket
 *
 *  \return unsigned int - number of bytes returned
 *
 */
unsigned int Net_Ethernet_UserUDP( UDP_Dsc* udpDsc )
{
    // Port 123 is a SNTP request
    if( udpDsc->destPort == SNTP_PORT )
    {
        SNTP_Response_t tmResp; //!< Holds packet framework for SNTP servers
        uint32_t tts = 0;       //!< Holds the time returned from the remote SNTP
        int ntpCount;

        tmResp.serverFlags = Net_Ethernet_getByte();
        tmResp.serverStratum = Net_Ethernet_getByte();
        // skip poll
        Net_Ethernet_getByte();
        tmResp.serverPrecision = Net_Ethernet_getByte();

        // skip all unused fields
        for( ntpCount = 0 ; ntpCount < 36 ; ntpCount++ )
        {
            Net_Ethernet_getByte();
        }

        // store transmit timestamp
        Highest( tts ) = Net_Ethernet_getByte();
        Higher( tts )  = Net_Ethernet_getByte();
        Hi( tts )      = Net_Ethernet_getByte();
        Lo( tts )      = Net_Ethernet_getByte();
        // convert sntp timestamp to unix epoch
        // The large number is the difference between 1900 and 1970
        tmResp.epoch = ( tts - 2208988800 );

        ntp_epoch_time = tmResp.epoch;

        update_GMT_str( GMT_string );

        ntp_received( ntp_epoch_time );

        return ( udpDsc->dataLength );
    }

    return 0;
}

/**********************************
 *
 *     Getters
 *
 **********************************/
// Retrieves SNTP timestamp from input
int8_t webserver_request_NTP( char* const ntpSrv )
{
    char* p_remoteIP = NULL;
    int8_t tries   = 5;

    do
    {
        debug_print( "Wait... Trying to resolve NTP server\r\n" );
        p_remoteIP = Net_Ethernet_dnsResolve( ntpSrv, 5 );
    }
    while( ( p_remoteIP == NULL ) && ( tries-- > 0 ) );

    if( tries > 0 )
    {
        uint8_t remoteIpAddr[4];
        uint8_t sntpPkt[48];

        memcpy( remoteIpAddr, p_remoteIP, 4 );
        // FLAGS : byte 0
        sntpPkt[0] = 0b00011001; // STRATUM : byte 1 = 0 - LI = 0 ; VN = 3 ; MODE = 1
        sntpPkt[1] = 0x00;       // POLL : byte 2
        sntpPkt[2] = 0x0a;   // PRECISION : byte 3 - 1024 sec (arbitrary value)
        sntpPkt[3] = 0xfa;             // 0.015625 sec (arbitrary value)
        // DELAY : bytes 4 to 7 = 0.2656 sec (arbitrary value)
        sntpPkt[6] = 0x44 ;
        // DISPERSION : bytes 8 to 11 = 16 sec (arbitrary value)
        sntpPkt[9] = 0x10 ;
        // REFERENCE ID : bytes 12 to 15 = 0 (unspecified)
        // REFERENCE TIMESTAMP : bytes 16 to 23 (unspecified)
        // ORIGINATE TIMESTAMP : bytes 24 to 31 (unspecified)
        // RECEIVE TIMESTAMP : bytes 32 to 39 (unspecified)
        // TRANSMIT TIMESTAMP : bytes 40 to 47 (unspecified)

        Net_Ethernet_sendUDP( remoteIpAddr, SNTP_PORT, SNTP_PORT, sntpPkt, 48 );
        return 0;
    }

    else
    {
        UART1_Write_Text( "Could not resolve DNS address for Time\n" );
        return -1;
    }
}

/*!
 *  \brief Requests a file from a remote server
 *
 *  \param char* - request Needs to be formatted http://www.happyserver.com/myfile.pdf etc...
 *  \param uint16_t - port number of port to be used on request.
 *
 *  \return Success of resolution and generation of request
 *    \retval 0 - Success
 *    \retval -1 - Failed
 *
 */
int8_t webserver_request_file( char* const request, uint16_t port )
{

    if( server_request.eth_socket_download != NULL )
    {
        debug_print( "Download in progress\r\n" );
        return -1;
    }

    else
    {
        // Check to see if the request is formatted correctly, must begin with http://
        if( strncmp( request, "http://", 7 ) != 0 )
        {
            debug_print( "Not a http request\r\n" );
            return -1;
        }

        else
        {
            char* p_remoteIP     = NULL;
            char* p_request      = NULL;
            int8_t tries         = 5;
            int domainCount      = 0;
            char domain[DOMAINSIZE];

            p_request = strstr( request, "//" );  // Forward pointer past the http:

            p_request += 2;                       // Now forward 2 more to get ahead of the //

            while( *p_request != '/' && domainCount < DOMAINSIZE )  // Past the // is the domain lets get it!
            {
                domain[domainCount] = *( p_request++ );
                domainCount++;
            }

            // Just incase, lets add that termination char
            domain[domainCount] = '\0';

            // TODO: if this is the same as last time, no need to resolve, speed things up
            do
            {
                debug_print( "Wait... Trying to resolve DNS\r\n" );
                p_remoteIP = Net_Ethernet_dnsResolve( domain, 5 );
            }
            while( ( p_remoteIP == NULL ) && ( tries-- > 0 ) );

            if( tries > 0 )
            {
                char remoteIPAddr[4] = {0};
                socket_t* client_socket;
                char* p_buffer;
                int i;

                client_socket = stack_pop( socketStack );

                p_buffer = client_socket->buffer_large;

                strcpy      ( p_buffer, "GET " );
                strcat      ( p_buffer, p_request );
                strcat_const( p_buffer, http_request_host );
                strcat      ( p_buffer, domain );
                strcat_const( p_buffer, http_request_end );

                for( i = 0; request[i]; i++ )
                {
                    if( request[i] == '/' )
                    {
                        if( request[i+1] )
                        {
                            p_request = &request[i+1];
                        }
                    }
                }

                if( strchr( p_request, '.' ) != 0 )
                {
                    strcpy( client_socket->file_name, p_request );
                }

                tries = 5;
                memcpy( remoteIPAddr, p_remoteIP, 4 );

                server_request.transmit_size     = strlen( p_buffer );
                client_socket->socketSM          = SM_HTTP_SVR_SEND_REQUEST;
                server_request.request_send_flag = TRUE;

                while( ( Net_Ethernet_connectTCP( remoteIPAddr, port, REQUESTPORT, &server_request.eth_socket_download ) != 1 )
                        && ( tries-- > 0 ) ) ;

                if( tries > 0 )
                {
                    client_socket->eth_socket = server_request.eth_socket_download;
                    return 0;
                }

                else
                {
                    close_socket( client_socket );
                    debug_print( "Could not connect\r\n" );
                    return -1;
                }

            }

            else
            {
                debug_print( "Could not resolve DNS\r\n" );
                return -1;
            }

        }

    }

}

/*!
 *  \brief Updates both the hardware timer as well as GMT timestamp
 *
 */
void webserver_time_update()
{
    int socketCount;

    for( socketCount = 0; socketCount < MAX_SOCKET; socketCount++ )
    {
        if( socketInUse[socketCount] != NULL )
        {
            socketInUse[socketCount]->alive_time++;
        }
    }

    Net_Ethernet_UserTimerSec++; /*<! Variable used with DHCP, DNS, and ARP functions */
    ntp_epoch_time++;            /*<! UNIX timestamp used for GMT time in headers */

}

/*!
 *  \brief Addes a static webpage to the server catalog
 *
 *  \param const char* page: pointer to static page
 *
 *  \return int8_t successful add or not
 *    \retval 0 = success, -1 = unsuccessful
 *
 *  \note
 *   <notes>
 */
int8_t webserver_add_staticpage( const char* page, char* page_name, long page_size )
{
    if( page_number == MAX_PAGES )
    {
        return -1;
    }

    else
    {
        char* p_file_extension;
        int extension_index;
        char* txt_buffer[12];

        // String representation of page
        strcpy( pages[page_number].page_name, page_name );
        // Get string of size
        LongToStr( page_size, txt_buffer );
        strcpy( pages[page_number].page_size_str, Ltrim( txt_buffer ) );
        // Forward to . extension
        p_file_extension = strchr( page_name, '.' );
        p_file_extension++;

        // Iterate through array
        for( extension_index = 0; extension_index < HTTP_UNKNOWN; extension_index++ )
        {
            if( strncmp_const( p_file_extension, http_file_extensions[extension_index], 3 ) == 0 )
            {
                pages[page_number].page_type = extension_index;
            }
        }

        // Set flag to represent a static page
        pages[page_number].static_page_flag = TRUE;
        // Set pointers
        pages[page_number].static_page.start_ptr = page;
        pages[page_number].static_page.current_ptr = page;
        pages[page_number].static_page.end_ptr = page + page_size;

        return page_number++;
    }
}

/*!
 *  \brief Addes a dynamic webpage to the server catalog
 *
 *  \param const char* page: pointer to static page
 *
 *  \return int8_t successful add or not
 *    \retval 0 = success, -1 = unsuccessful
 *
 *  \note
 *   <notes>
 */
int8_t webserver_add_dynamicpage( char* const page, char* page_name, void ( *page_upd_callback )( void ) )
{
    if( page_number == MAX_PAGES )
    {
        return -1;
    }

    else
    {
        char* p_file_extension;
        int extension_index;
        char txt_buffer[12];

        // String representation of page
        strcpy( pages[page_number].page_name, page_name );
        // Get string of size
        LongToStr( strlen( page ), txt_buffer );
        strcpy( pages[page_number].page_size_str, Ltrim( txt_buffer ) );
        // Forward to . extension
        p_file_extension = strchr( page_name, '.' );
        p_file_extension++;

        // Iterate through array
        for( extension_index = 0; extension_index < HTTP_UNKNOWN; extension_index++ )
        {
            if( strncmp_const( p_file_extension, http_file_extensions[extension_index], 3 ) == 0 )
            {
                pages[page_number].page_type = extension_index;
            }
        }

        // Set flag to represent a static page
        pages[page_number].static_page_flag = FALSE;
        // Set pointers
        pages[page_number].dynamic_page.start_ptr = page;
        pages[page_number].dynamic_page.current_ptr = page;
        pages[page_number].dynamic_page.end_ptr = page + strlen( page );
        pages[page_number].dynamic_page.page_update = page_upd_callback;

        return page_number++;
    }
}


/*!
 *  \brief Initializes the webserver
 *
 *  \param int mode - Either 0 for no IRQ 1 for IRQ mode
 *
 */
int8_t webserver_init( struct webserver_config_t* config )
{

    http_port     = config->port;
    ntp_epoch_time= config->epochTime;
    form_received = config->form_key_values;
    file_download = config->file_download;
    ntp_received  = config->ntp_time;
    irqModeEnable = config->irq_enable;

    packet_flag   = TRUE;        /*<! Set packet_flag to TRUE.  If IRQ mode is disabled, this will keep it in a loop */

    Net_Ethernet_stackInitTCP(); /*<! Initialize TCP stack, must be done first */

#ifdef AVR

    /*while( !( Net_Ethernet_28j60_readReg( ESTAT ) & 0x01 ) )
    {
        Delay_ms( 200 );
        debug_print( "Waiting for ethernet clock\r\n" );
    }*/

    Net_Ethernet_Init( config->myMacAddr, config->myIpAddr, Net_Ethernet_28j60_FULLDUPLEX );

    if( irqModeEnable == TRUE )
    {
        Net_Ethernet_28j60_INT_Direction = 0;

        EICRA |= ( 1<<ISC11 );                   // Enable Falling Edges
        EIMSK |= ( 1<<INT1 );                    // Enable external pin IRQ
        EIFR &= ~( 1<<INTF1 );                   // Clear IRQ Flag

        Net_Ethernet_28j60_setBitReg( EIE, INTIE | PKTIE );

    }

#endif

#ifdef STM32

    if( Net_Ethernet_Init( config->myMacAddr,
                           config->myIpAddr,
                           _ETHERNET_AUTO_NEGOTIATION | _ETHERNET_FULLDUPLEX,
                           &_GPIO_MODULE_ETHERNET ) == 1 )
    {
        return -1;
    }

#endif

#ifdef STM32PLUS

    if( Net_Ethernet_Init( config->myMacAddr, config->myIpAddr,
                           _ETHERNET_FULLDUPLEX | _ETHERNET_SPD10,
                           &_GPIO_MODULE_ETHERNET_G11_13_14 ) == 1 )
    {
        return -1;
    }

#endif

    Net_Ethernet_confNetwork( config->ipMask, config->gwIpAddr, config->dnsIpAddr );

#ifndef AVR

    if( irqModeEnable == TRUE )
    {
        // Wake up IRQ on EXTI19
        Ethernet_DMA_DMAIER |= ( 1 << NISE ) | ( 1 << RIE );
        NVIC_IntEnable( IVT_INT_ETH );      /*!< Enable ethernet interrupts */
    }

    {
        int regValue = 0;
        const char _PHY_BSR = 1;

        // Wait for active link
        while( !( regValue & 0b100 ) )
        {
            Net_Ethernet_Intern_PHY_Read( _PHY_BSR, &regValue );
        }
    }

    // Asked not to use in Errata document
    //Ethernet_DMA_DMAOMR |= ( 1 << FTF );    /*!< Flush transmit fifo */

    //dma_mem_init( CHANNEL_1, DMA_HGH_PRIORITY );
    //Ethernet_DMA_DMAOMR &= ~( 1 << 25 ) | ( 1 << 21 );
#endif
    {
        int socketCount;
        socket_t* soc;

        // Populate available stockets
        for( socketCount = 0; socketCount < MAX_SOCKET; socketCount++ )
        {
            soc = &sockets[socketCount];
            stack_push( socketStack, soc );
        }

        server_request.eth_socket_download = NULL;
        server_request.request_send_flag   = FALSE;
        server_request.capture_length      = 0;

    }

    return 0;
}
/*
 *   tutorial5solution.htm?check0=on&text0=Richard&check1=on&text1=Lowe&check2=on&text2=ocdrichard%40gmail.com&check3=on&text3=05%2F25%2F1974
 */

/**********************************
 *
 *      Tasks
 *
 **********************************/
/*!
 *  \brief Cleans up closed sockets
 *
 */
void webserver_cleanup()
{
    int socketCount;

    for( socketCount = 0; socketCount < MAX_SOCKET; socketCount++ )
    {
        if( socketInUse[socketCount] != NULL )
        {
            if( ( socketInUse[socketCount]->eth_socket->state == 0 )
                    || ( socketInUse[socketCount]->alive_time > MAX_ALIVE_TIME ) )
            {
                close_socket( socketInUse[socketCount] );
            }
        }
    }

    update_GMT_str( &GMT_string );

}

/*!
 *  \brief Processes incoming packets
 *
 */
void webserver_read_packet()
{
    Net_Ethernet_doPacket();

    if( ( server_request.eth_socket_download->state == 3 ) && ( server_request.request_send_flag == TRUE ) )
    {
        if( Net_Ethernet_startSendTCP( server_request.eth_socket_download ) == 1 )
        {
            server_request.request_send_flag = FALSE;
        }
    }

#ifdef STM32
    bar_graph_display( ( MAX_SOCKET-1 ) - stackTop );
#endif

}

/*!
 *  \brief Used for debugging purposes
 *
 */
#ifdef DEBUG
void webserver_print_status( void )
{
    char txtBuffer[50];
    int count;
    socket_t* mySocket;

    if( server_request.eth_socket_download != NULL )
    {
        sprinti( txtBuffer, "***\r\n\nState of download socket: %u ***\r\n", ( int )server_request.eth_socket_download->state );
        debug_print( txtBuffer );
    }

}
#endif

/*************************************
 *    ISRs
 *************************************/
// Ethernet IRQ
void packet_ISR() iv packetVector ics ICS_AUTO
{
#if AVR
    /* As per page 65 in the enc28j60 datasheet
       The interrupt pin is designed for use by a host controller that is
       capable of detecting falling edges.*/

    /* Step 1. After an interrupt occures, the host should clear the global
       enable bit for the interrupt pin before servicing the interrupt.
       Clearing the enable bit will cause the interrupt pin to return to the
       non-asserted state (high) */
    Net_Ethernet_28j60_clearBitReg( EIE, INTIE ); // Address , Mask

    if( ( Net_Ethernet_28j60_readReg( EIR ) & PKTIF ) > 0 )
    {
        packet_flag = TRUE;
    }

    /* Step 1a. After the interrupt has been serviced, the global enable bit
       may be restored */
    //Net_Ethernet_28j60_setBitReg( EIR, INTIE );

    /* NOTE: Except for the LINKIF interrupt flag, interrupt flag bits are set
       when an interrupt condition occurs regardless of the state of its
       corresponding enable bit or the associated global enable bit. User
       software should ensure the appropriate interrupt flag bits are clear
       prior to enabling an interrupt. This feature allows for software
       polling. */

#else

    // Normal interrupt summary
    if( Ethernet_DMA_DMASR & ( 1 << NIS ) )
    {
        // Bit 6 - Reveive interrupt
        if( Ethernet_DMA_DMASR & ( 1 << RS ) )
        {
            packet_flag = TRUE;
            Ethernet_DMA_DMASR |= ( 1 << RS );
        }

        Ethernet_DMA_DMASR |= ( 1 << NIS );
    }

#endif
}

/*#ifndef AVR
void transferDone_ISR() iv IVT_INT_DMA2_Stream1 ics ICS_AUTO
{
    dma_mem_complete();
}
#endif*/